// Receiver for HiSKY H-6 Transmitter
// 2016-11-23  T. Nakagawa

#include <Servo.h>
#include "NRF24L01.h"

#define SERVO_NUM 4
#define SERVO_PINS {5, 6, 7, 8}
#define NRF24L01_PIN_CE 10
#define NRF24L01_PIN_CS 9
#define BIND_CHANNEL 81
#define BIND_ADDRESS ((const uint8_t *)"\x12\x23\x23\x45\x78")
#define PACKET_SIZE 10
#define PULSE_MIN (1500 - 500)
#define PULSE_MAX (1500 + 500)

Servo servos[SERVO_NUM];
static uint8_t data_address[5];
static uint8_t data_channels[20];
NRF24L01 radio(NRF24L01_PIN_CE, NRF24L01_PIN_CS);

void setup() {
  const uint8_t servo_pins[SERVO_NUM] = SERVO_PINS;
  for (uint8_t i = 0; i < SERVO_NUM; i++) {
    servos[i].attach(servo_pins[i]);
  }

  radio.setCRC(NRF24L01::CRC_2BYTE);
  radio.setDataRate(NRF24L01::DR_1MBPS);
  radio.setChannel(BIND_CHANNEL);
  radio.setPipes(1);
  radio.setAddressWidth(5);
  radio.setAddress(0, BIND_ADDRESS);
  radio.setPayloadSize(PACKET_SIZE);
  radio.enable();
}

void loop() {
  static uint8_t state = 0;  // 0-3: Binding, 4: Data receiving.
  static union {uint16_t uint16; uint8_t uint8[2];} adrs_sum = {0};
  static uint8_t hop = 0;
  static uint32_t timeout;
  static uint8_t retry;

  if (state > 4 && retry < 5 && micros() >= timeout) {
    radio.disable();
    radio.setChannel(data_channels[hop++]);
    if (hop >= 20) hop = 0;
    radio.enable();
    timeout = micros() + 10000UL;
    retry++;
  }

  while (!radio.empty()) {
    uint8_t buf[PACKET_SIZE];
    radio.read(buf);
    if (state == 0) {
      if (buf[0] == 0xff && buf[1] == 0xaa && buf[2] == 0x55) {
        for (uint8_t i = 0; i < 5; i++) {
          adrs_sum.uint16 += (data_address[i] = buf[3 + i]);
        }
        state++;
      }
    } else if (state < 4) {
      if (buf[0] == adrs_sum.uint8[0] && buf[1] == adrs_sum.uint8[1] && buf[2] == state - 1) {
        for (uint8_t i = 0; i < 7; i++) {
          data_channels[7 * (state - 1) + i] = buf[3 + i];
          if (state == 3 && i == 5) {
            radio.disable();
            radio.setAddress(0, data_address);
            radio.setChannel(data_channels[hop++]);
            radio.enable();
            timeout = micros() + 10000UL;
            retry = 0;
            break;
          }
        }
        state++;
      }
    } else {
      radio.disable();
      radio.setChannel(data_channels[hop++]);
      if (hop >= 20) hop = 0;
      radio.enable();
      timeout = micros() + 10000UL;
      retry = 0;

      uint16_t high = *(uint16_t *)(buf + 8);
      for (int8_t i = 0; i < 8; i++) {
        uint16_t val = (((high & 0x0003) << 8) | buf[i]);
        val = constrain(map(val, 150, 1000 - 150, PULSE_MIN, PULSE_MAX), PULSE_MIN, PULSE_MAX);
        high >>= 2;
        if (i < SERVO_NUM) {
          servos[i].writeMicroseconds(val);
        }
      }
    }
  }
}

